/* An OrderedMap represents a Map which allows array-style indexing over the keys
 * While the normal map keeps insertion order, this gives an easier way to access arbitrary indicies without using iterators */
class OrderedMap {
    constructor() {
        this.map = new Map();
        this.order = [];
    }
    append(key, value) {
        this.map.set(key, value);
        this.order.push(key);
    }
    prepend(key, value) {
        this.map.set(key, value);
        this.order.unshift(key);
    }
    has(key) {
        return this.map.has(key);
    }
    get(key) {
        return this.map.get(key);
    }
    getAt(index) {
        return this.map.get(this.order[index]);
    }
    getPairAt(index) {
        const mapKey = this.order[index];
        return [mapKey, this.map.get(mapKey)];
    }
    getKeyAt(index) {
        return this.order[index];
    }
    getKeys() {
        return this.order;
    }
    forEach(fn) {
        this.map.forEach(fn);
    }
    indexOf(key) {
        return this.order.indexOf(key);
    }
    delete(key) {
        const index = this.order.indexOf(key);
        if (index === -1)
            return false;
        this.order.splice(index, 1);
        return this.map.delete(key);
    }
    deleteAt(index) {
        const mapKey = this.order[index];
        this.order.splice(index, 1);
        return this.map.delete(mapKey);
    }
    clear() {
        this.map.clear();
        this.order = [];
    }
    get length() {
        return this.order.length;
    }
}

/* Gets the intersection of the Array type (not Set)
 * Implementation assumes that array items are unique */
function getArrayIntersection(...sets) {
    const numOfAppearances = new Map();
    sets.forEach(set => {
        set.forEach(item => {
            if (!numOfAppearances.has(item))
                numOfAppearances.set(item, 0);
            const incrementedValue = numOfAppearances.get(item) + 1;
            numOfAppearances.set(item, incrementedValue);
        });
    });
    const intersection = [];
    numOfAppearances.forEach((count, item) => {
        if (count === sets.length)
            intersection.push(item);
    });
    return intersection;
}
class MultidimensionalMap {
    constructor(dimensions, entries) {
        this.entries = [];
        this.dimensions = dimensions.reduce((acc, curr) => {
            acc[curr] = new OrderedMap();
            return acc;
        }, {});
        if (entries) {
            this.addEntries(entries);
        }
    }
    addEntries(entries) {
        entries.forEach(entry => {
            this.entries.push(entry);
            Object.keys(this.dimensions).forEach(dimension => {
                const entryValueForDimension = entry[dimension];
                const dimensionMap = this.dimensions[dimension];
                if (!dimensionMap.has(entryValueForDimension)) {
                    dimensionMap.append(entryValueForDimension, []);
                }
                dimensionMap.get(entryValueForDimension).push(entry);
            });
        });
    }
    getAllEntries() {
        return this.entries;
    }
    getEntriesInRange(dimension, start, end) {
        if (!this.dimensions.hasOwnProperty(dimension))
            throw new Error(`Dimension "${dimension}" does not exist`);
        const targetDimension = this.dimensions[dimension];
        const startIdx = start == null ? 0 : targetDimension.indexOf(start);
        const endIdx = end == null ? targetDimension.length - 1 : targetDimension.indexOf(end);
        if (startIdx === -1)
            throw new Error(`Range start "${start}" does not exist in dimension`);
        if (endIdx === -1)
            throw new Error(`Range end "${end}" does not exist`);
        const entryList = [];
        for (let i = startIdx; i <= endIdx; i++) {
            const entries = targetDimension.getAt(i);
            entryList.push(...entries);
        }
        return entryList;
    }
    getSubsetArray(query) {
        const subsets = Object.entries(query).map(([dimensionName, dimensionItem]) => {
            if (!this.dimensions.hasOwnProperty(dimensionName))
                throw new Error(`Dimension "${dimensionName}" does not exist`);
            if (typeof dimensionItem === 'object') {
                if (dimensionItem.range && dimensionItem.matches) {
                    throw new Error(`Must specify either "range" or "matches" property but not both`);
                }
                else if (dimensionItem.range) {
                    const [start, end] = dimensionItem.range;
                    return this.getEntriesInRange(dimensionName, start, end);
                }
                else if (dimensionItem.matches) {
                    const matchingEntries = [];
                    dimensionItem.matches.forEach((queryItem) => {
                        const matchResult = this.dimensions[dimensionName].get(queryItem);
                        matchingEntries.push(...matchResult);
                    });
                    return matchingEntries;
                }
                else {
                    return this.entries;
                }
            }
            return this.dimensions[dimensionName].get(dimensionItem);
        });
        return getArrayIntersection(...subsets);
    }
    getSubset(query) {
        const subsetArray = this.getSubsetArray(query);
        return new MultidimensionalMap(Object.keys(this.dimensions), subsetArray);
    }
    combineEntries(measure, fields, entries) {
        const dataEntries = entries ? entries : this.entries;
        if (dataEntries.length === 0)
            return [];
        const nestedEntries = {};
        if (fields == null || fields.length === 0) {
            nestedEntries[measure] = 0;
            dataEntries.forEach(entry => nestedEntries[measure] += entry[measure]);
            return nestedEntries;
        }
        fields.forEach(field => {
            if (!this.dimensions.hasOwnProperty(field))
                throw new Error(`Field "${field}" does not exist`);
        });
        dataEntries.forEach(dataEntry => {
            let current = nestedEntries;
            fields.forEach((subfield, idx) => {
                if (current[dataEntry[subfield]] == null) {
                    if (idx >= fields.length - 1) {
                        const subset = { [measure]: dataEntry[measure] };
                        fields.forEach(subfield => { subset[subfield] = dataEntry[subfield]; });
                        current[dataEntry[subfield]] = subset;
                    }
                    else {
                        current[dataEntry[subfield]] = {};
                    }
                }
                else {
                    if (idx >= fields.length - 1) {
                        current[dataEntry[subfield]][measure] += dataEntry[measure];
                    }
                }
                current = current[dataEntry[subfield]];
            });
        });
        return nestedEntries;
    }
    get length() {
        return this.entries.length;
    }
}

export { MultidimensionalMap };
